from random import randint, gauss, shuffle, choices
from datetime import datetime, timedelta
import pickle

from sudoku_py import SudokuGenerator, Sudoku

from rollbot import as_plugin, with_help, with_startup, as_sender_singleton, as_group_singleton, RollbotFailure


@as_sender_singleton
class RollcoinWallet:
    balance: float = 0.0


@as_sender_singleton
class RollcoinBegCooldown:
    cooldown: datetime = datetime(2000, 1, 1)


@as_group_singleton
class RollcoinBlockchain:
    puzzle: "binary"

    def get_puzzle(self):
        if self.puzzle is None:
            return None
        return pickle.loads(self.puzzle)

    def set_puzzle(self, new_puzzle):
        self.puzzle = pickle.dumps(new_puzzle)
    
    def clear_puzzle(self):
        self.puzzle = None


@as_sender_singleton
class RollcoinGamblingWallet:
    balance: float = 0.0


@as_group_singleton
class RollcoinMarket:
    state: str = "neutral"


def setup_initial_balances(plugin, db):
    bot = plugin.bot
    bot.logger.info("Setting up rollcoin wallets")
    wallets = bot.config.get("rollcoin.wallets")
    initial = bot.config.get("rollcoin.initial")
    force = bot.config.get("rollcoin.force_initial")
    for k, v in initial.items():
        wallet_key = wallets.get(k, None)
        if wallet_key is None:
            continue
        wallet = db.query(RollcoinWallet).get(wallet_key)
        if wallet is None:
            wallet = RollcoinWallet.create_from_key(wallet_key)
        elif k not in force:
            continue
        wallet.balance = v
        db.add(wallet)
    bot.logger.info("Initial wallets created!")


@with_startup(setup_initial_balances)
@with_help("Check your balance of Rollcoins")
@as_plugin
def balance(wallet_data: RollcoinWallet, gambling_data: RollcoinGamblingWallet):
    return f"You have {wallet_data.balance} Rollcoins in your wallet and {gambling_data.balance} Rollcoins in the market"


@with_help("Tip someone else some Rollcoins: !tip target amount")
@as_plugin
def tip(msg, db, sender_wallet_data: RollcoinWallet, wallet_cfg: "rollcoin.wallets"):
    args = msg.arg_list()
    if len(args) != 2:
        return RollbotFailure.INVALID_ARGUMENTS.with_reason("Tip requires exactly 2 arguments - target and amount")
    
    target, raw_amount = args
    try:
        amount = float(raw_amount)
    except ValueError:
        return RollbotFailure.INVALID_ARGUMENTS.with_reason(f"Amount should be a number - not {raw_amount}")

    if amount <= 0:
        return RollbotFailure.INVALID_ARGUMENTS.with_reason(f"Amount should be positive")

    target_id = wallet_cfg.get(target.lower(), None)
    if target_id is None:
        return RollbotFailure.INVALID_ARGUMENTS.with_reason(f"Could not find wallet-holder {target}")

    if sender_wallet_data.balance < amount:
        return f"Insufficient funds! You need {round(amount - sender_wallet_data.balance, 2)} more Rollcoins. Have you tried to !mine, !beg, or !gamble?"
    
    target_wallet = db.query(RollcoinWallet).get(target_id)
    if target_wallet is None:
        target_wallet = RollcoinWallet.create_from_key(target_id)
        db.add(target_wallet)

    sender_wallet_data.balance = sender_wallet_data.balance - amount
    target_wallet.balance = target_wallet.balance + amount
    return f"Done! {target} now has {target_wallet.balance} Rollcoins"


@with_help("Receive a mining question you can answer to receive Rollcoins")
@as_plugin
def mine(msg, blockchain_data: RollcoinBlockchain, sender_data: RollcoinWallet):
    puzzle = blockchain_data.get_puzzle()
    if puzzle is None:
        exchange = list(zip("abcdefghi", range(10)))
        shuffle(exchange)
        gen = SudokuGenerator(9)
        gen.generate(0)
        gen.board_exchange_values({k: v + 1 for k, v in exchange})
        gen.generate(randint(10, 70))
        puzzle = gen.board
        blockchain_data.set_puzzle(puzzle)

    if msg.raw_args is None:
        return "The current mining challenge is\n" + str(Sudoku(board=puzzle))

    parsed = []
    row = []
    for c in msg.raw_args:
        if not c.isdigit():
            continue
        row.append(int(c))
        if len(row) == 9:
            parsed.append(row)
            row = []
    
    try:
        guess = Sudoku(board=parsed, block_width=3, block_height=3)
    except:
        return "Sorry, I could not parse that solution!"

    for i, row in enumerate(puzzle):
        for j, cell in enumerate(row):
            if cell not in (0, parsed[i][j]):
                return "Sorry, that solution doesn't match the puzzle! The current mining challenge is\n" + str(Sudoku(board=puzzle))
    
    for row in parsed:
        if set(row) != set(range(1, 10)):
            return "Sorry, that solution isn't valid!"
    
    for i in range(9):
        if set(r[i] for r in parsed) != set(range(1, 10)):
            return "Sorry, that solution isn't valid!"

    for i in range(0, 9, 3):
        for j in range(0, 9, 3):
            if set(x for r in parsed[i:i+3] for x in r[j:j+3]) != set(range(1, 10)):
                return "Sorry, that solution isn't valid!"

    market_rate = round(abs(gauss(10, 1)), 2)
    sender_data.balance = sender_data.balance + market_rate
    blockchain_data.clear_puzzle()
    return f"Looks right to me! The current market rate is {market_rate} Rollcoins per Sudoku, so that brings your balance to {sender_data.balance}"


@with_help("Beg for some Rollcoins")
@as_plugin
def beg(wallet_data: RollcoinWallet, cooldown_data: RollcoinBegCooldown, cooldown_cfg: "rollcoin.cooldown", beg_chance_cfg: "rollcoin.beg_chance"):
    old_cooldown = cooldown_data.cooldown
    cooldown_data.cooldown = datetime.now() + timedelta(**cooldown_cfg)

    if datetime.now() < old_cooldown:
        return f"Don't beg so frequently - it's not a good look"

    if randint(0, 99) > beg_chance_cfg:
        return "What are you, poor?"

    increase = round(abs(gauss(1, 1)), 2)
    wallet_data.balance = wallet_data.balance + increase
    return f"Fine, I've given you {increase} Rollcoins, which brings your balance to {wallet_data.balance}"


@with_help("Gamble some Rollcoins")
@as_plugin
def gamble(db, msg, wallet_data: RollcoinWallet, gambling_data: RollcoinGamblingWallet, market_data: RollcoinMarket, market_cfg: "rollcoin.market"):
    # pos moves money wallet -> gambling, neg moves money gambling -> wallet
    # market has a current state (neutral, bull, bear, boom, bust, collapse, etc.)
    # after a transfer is made, the market evolves based on
    #   - direction of transfer (buy vs sell)
    #   - current market state
    #   - rng
    # market enters new state and affects all gambling wallets, multiplying their balances
    args = msg.arg_list()
    if len(args) != 1:
        return RollbotFailure.INVALID_ARGUMENTS.with_reason("Gambling requires exactly one argument: the amount to transfer (can be positive or negative)")

    try:
        amount = float(args[0])
    except ValueError:
        return RollbotFailure.INVALID_ARGUMENTS.with_reason(f"Amount should be a number - not {args[0]}")

    if amount == 0:
        return RollbotFailure.INVALID_ARGUMENTS.with_reason(f"Amount should be nonzero")

    buying = amount > 0
    abs_amount = abs(amount)
    if buying and abs_amount > wallet_data.balance:
        return f"Insufficient funds! You need {round(abs_amount - wallet_data.balance, 2)} more Rollcoins. Have you tried to !mine, !beg, or !gamble?"
    elif not buying and abs_amount > gambling_data.balance:
        return f"Insufficient holdings! You would need {round(abs_amount - gambling_data.balance, 2)} more Rollcoins in the market."

    options = market_cfg[market_data.state]["buy" if buying else "sell"]
    transition = choices(options, weights=tuple(x["weight"] for x in options))[0]
    multiplier = transition["multiplier"]
    market_data.state = transition["new_state"]
    wallet_data.balance = wallet_data.balance - amount
    gambling_data.balance = multiplier * (gambling_data.balance + amount)

    for holdings in db.query(RollcoinGamblingWallet
                            ).filter(RollcoinGamblingWallet.sender_id != gambling_data.sender_id
                            ).all():
        holdings.balance = multiplier * holdings.balance

    return f"{transition['message']}\nYour current holdings: {gambling_data.balance}"